# -*- coding: utf-8 -*-
### BEGIN LICENSE
# Copyright (C) 2010 Mads Chr. Olesen <mchro@cs.aau.dk>, Kenneth Yrke Jørgensen <kyrke@cs.aau.dk>
#This program is free software: you can redistribute it and/or modify it 
#under the terms of the GNU General Public License version 3, as published 
#by the Free Software Foundation.
#
#This program is distributed in the hope that it will be useful, but 
#WITHOUT ANY WARRANTY; without even the implied warranties of 
#MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
#PURPOSE.  See the GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License along 
#with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import pyuppaal
import copy
import logging
import sys
logger = logging.getLogger('succgen')
from collections import defaultdict

from pyuppaal.ulp import lexer, parser

from opaal.model_parsers.state import State
import opaal.util
class UnknownVarTypeException(opaal.util.Error):
    pass
class VirtualMachineException(opaal.util.Error):
    pass
import util
from simplify import SimplifyModel

from opaal.lattices import LatticeVector

# which externs can we use?
# name => (module, class)
available_externs = {
        'AbstractCache': ('opaal.lattices.abstract_cache', 'AbstractCache'),
        'UnionBitVector': ('opaal.lattices.bitvector', 'UnionBitVector'),
        'IntersectionBitVector': ('opaal.lattices.bitvector', 'IntersectionBitVector'),
        'MaxInteger': ('opaal.lattices.integer_lattice', 'MaxInteger')
        }

class SuccessorGenerator:
    def __init__(self, model, pwlist,maxconstant, constant_overrides=dict()):
        model = SimplifyModel(model, constant_overrides).simplify()

	self.pwlist = pwlist
        self.model = model
        self.transitions = {}
        self.sync_transitions = {}
        self.templatename_to_idx = {}
        self.num_templates = len(self.model.templates)
        self.invariants = {}
        self.externs = {}
        self.maxconstant = maxconstant
        self.infconstant = maxconstant + 2 #bad name, it is used for inf + 1

        #mapping from template => location id => nice name
        self.location_labels = defaultdict(dict)

        lex = lexer.lexer
        pars = parser.Parser(model.declaration, lex)

        declvisitor = parser.DeclVisitor(pars)
        self.clocks = declvisitor.clocks

        self.constants = {}
        #constant overrides
        for (cident, cval) in constant_overrides.iteritems():
            #XXX assumes expression is already python
            self.constants[cident] = eval(cval, {}, self.constants)
        #evaluate constants to values
        for (cident, cval) in declvisitor.constants.iteritems():
            if not cident in self.constants:
                self.constants[cident] = eval(util.expression_to_python(cval), {}, self.constants)

        #evaluate array dimensions to values
        self.variables = []
        for (vident, vtype, vdimensions) in declvisitor.variables:
            for (dim, dim_idx) in zip(vdimensions, range(len(vdimensions))):
                vdimensions[dim_idx] = eval(util.expression_to_python(dim), {}, self.constants)
            self.variables += [(vident, vtype, vdimensions)]

        #evaluate channel array dimensions to values
        self.channels = []
        def crossproduct(vdimensions):
            if len(vdimensions) == 1:
                for i in xrange(vdimensions[0]):
                    yield "[%d]" % i
            else:
                for i in xrange(vdimensions[0]):
                    for t in crossproduct(vdimensions[1:]):
                        yield ("[%d]" % i) + t

        for (vident, vdimensions) in declvisitor.channels:
            if len(vdimensions) == 0:
                self.channels += [vident]
                continue
            for (dim, dim_idx) in zip(vdimensions, range(len(vdimensions))):
                vdimensions[dim_idx] = eval(util.expression_to_python(dim), {}, self.constants)
            for i in crossproduct(vdimensions):
                self.channels += [vident + i]

        #import needed externs
        for ex in pars.externList:
            modname, clsname = available_externs[ex]
            __import__(modname)
            mod = sys.modules[modname]
            cls = getattr(mod, clsname)
            self.externs[ex] = cls

        logger.info("Constants: %s", self.constants)
        logger.info("Variables: %s", self.variables)
        logger.info("Clocks: %s", self.clocks)
        logger.info("Externs: %s", self.externs)

        #Calculate invariants
        for (t, t_idx) in zip(self.model.templates,
                            range(len(self.model.templates))):
            self.invariants[t_idx] = {}
            for (loc, l_idx) in zip(t.locations,
                                range(len(t.locations))):
                if (loc.invariant.value != None):
                    self.invariants[t_idx][l_idx] = util.expression_to_python_DEPR(loc.invariant.value) #Set inv
                else:
                    self.invariants[t_idx][l_idx] = ''

        #calculate transitions
        for (t, t_idx) in zip(self.model.templates, 
                            range(len(self.model.templates))):
            self.transitions[t_idx] = {}
            self.sync_transitions[t_idx] = {}
            self.templatename_to_idx[t.name] = t_idx

            for (loc, l_idx) in zip(t.locations,
                                range(len(t.locations))):
                if loc.name.get_value():
                    self.location_labels[t_idx][l_idx] = loc.name.get_value()

                self.transitions[t_idx][l_idx] = []
                self.sync_transitions[t_idx][l_idx] = {}
                for chan_name in self.channels:
                    self.sync_transitions[t_idx][l_idx][chan_name + "!"] = []
                    self.sync_transitions[t_idx][l_idx][chan_name + "?"] = []
                for tr in [tr for tr in t.transitions if tr.source == loc]:
                    target_idx = t.locations.index(tr.target)
                    debug_info = {}

                    lowerVal = {}
                    higherVal = {}
                    resets = set()
                    guard_code = None
                    debug_info['guard_code'] = ''
                    if tr.guard.value != '':
			#print tr.guard.value
                        
			elements = tr.guard.value.split("and")
			for s in elements:
				splitGuard = s.strip().split(" ")

				if splitGuard[1] == ">=":
                                    lowerVal[splitGuard[0]] = int(splitGuard[2])
                                if splitGuard[1] == ">":
                                    lowerVal[splitGuard[0]] = int(splitGuard[2])+1

                                if splitGuard[1] == "<=":
				    higherVal[splitGuard[0]] = int(splitGuard[2])
                                if splitGuard[1] == "<":
				    higherVal[splitGuard[0]] = int(splitGuard[2])-1

				if splitGuard[1] == "==":
				    lowerVal[splitGuard[0]] = int(splitGuard[2])
				    higherVal[splitGuard[0]] = int(splitGuard[2])

                        #logger.debug("Guard: %s ==> %s", tr.guard.value, util.expression_to_python_DEPR(tr.guard.value))
                        guard_code = compile(util.expression_to_python_DEPR(tr.guard.value), "<guard>", 'eval')
                        debug_info['guard_code'] = tr.guard.value

                    update_code = None
                    debug_info['update_code'] = ''
                    if tr.assignment.value != '':
                        update_code = compile(util.expression_to_python_DEPR(tr.assignment.value), "<update>", 'exec')
                        debug_info['update_code'] = tr.assignment.value

                    list_curtrans = [(target_idx, guard_code, update_code, 
                        tr.synchronisation.value, debug_info, lowerVal, higherVal, resets)]

                    self.transitions[t_idx][l_idx] += list_curtrans

                    if tr.synchronisation.value != "":
                        self.sync_transitions[t_idx][l_idx][tr.synchronisation.value] += list_curtrans

        #print self.location_labels
        #print "Transitions:"
        #print self.transitions

#    def checkInvariant(self, a):
#        state,_ = a
#        for t_idx in self.invariants.keys():
#            cur_inv = self.invariants[t_idx][state.locs[t_idx]]
#            if cur_inv != '' and not eval(cur_inv, self.constants, state):
#                return False
#        return True


    def normalize(self, state, w, d):
        m = self.maxconstant+1
        for c,_ in self.clocks:
            tmp = w+state[c]
            mintmp = min([w+state[c], self.maxconstant+1])
            state[c] = mintmp
            if mintmp <= m:
                m = state[c]

        if w != m: #No change in the anchor point
	    for c,_ in self.clocks:
	        state[c] = state[c] - m

	return (state, m, self.infconstant)

    def trans_successors(self, state, w, p, trans_info=False):
        #Take an active transition
        for t_idx in xrange(self.num_templates):
            for (target_idx, guard, update, sync, debug_info, lowerVal, higherVal, resets) in self.transitions[t_idx][state.locs[t_idx]]:
                #logger.debug("")
                #logger.debug("")
                #Evaluate guard

                #logger.debug("New state")
                #logger.debug("%s - %s - %s", state, w, p)
                
                #Find min and max delay for each clocks 
#                start = 0
#                for c,_ in self.clocks:
#                    guard = 0
#                    try:
#                        guard = lowerVal[c] - state[c]
#                    except Exception, e:
#                        continue #No lower guard set, is enabled from now
#                    if guard > start:
#                        start = guard
#        
#                end = self.maxconstant + 1
#                for c,_ in self.clocks:
#                    guard = 0
#                    try:
#                        guard = higherVal[c] - state[c]
#                    except Exception, e:
#                        continue #No high guard set, assuming the highest possible
#        
#                    if guard < end:
#                        end = guard
                 
#Faster than the above if we have few guards
#If the mode always has a upper and lower bound making the union of
#lowerVal.keys and higerVal.keys might be faster
                start = 0
                for c in lowerVal.keys():
                    guard = lowerVal[c] - state[c]
                    if guard > start:
                        start = guard

                if start < 0 or p <= start: #Terminate early if can be satisfyed or if already passed
                    return

                end = self.maxconstant + 1
                for c in higherVal.keys():
                    guard = higherVal[c] - state[c]
                    if guard < end:
                        end = guard
                 

# XXX: I would expect the following to be faster, however it seems like its
# important to use continue, i guess it is faster if all clocks has a guard
#                start = 0
#                end = self.maxconstant + 1
#                for c, _ in self.clocks:
#                    guardLo = 0
#                    guardHi = self.infconstant
#                    clockVal = state[c]
#                    try:
#                        guardLo =  lowerVal[c] - clockVal
#                    except LookupError, e:
#                        pass
#                    
#                    if guardLo > start:
#                        start = guardLo
#
#                    try:
#                        guardHi = higherVal[c] - clockVal
#                    except LookupError, e:
#                        pass 
#                                        
#                    if guardHi < end:
#                        end = guardHi

                
                #logger.debug("lowerVals %s, higerVals %s", lowerVal, higherVal)
                #logger.debug("The state has start %s and end %s ", start, end)
        
                #Check if the guard can be satisfied
                if  start > end or w > end:
                    return
        
                #Calculate discrete update
                newState = state.copy()
                newState.locs[t_idx] = target_idx
                #Calculate successor states
                #logger.debug("Successors:")
                if update == None:
                    for c, _ in self.clocks:
                        newVal = newState[c] + w
                        if newVal > self.maxconstant + 1: #Normalize if larger than max constant 
                            newVal = self.maxconstant + 1
                        newState[c] = newVal - w
                        
                    yield (newState, w, self.infconstant) # We only need to normalize if we had a open higher guard
                else: #Is allready passed?
                    if p < end:
                        end = p - 1 # Minus one as value p is passed
                    #yield many states
                    for i in range(start, end+1): #dont need infin as max is mc+1
                        a = newState.copy()
                        for c,_ in self.clocks:
                            tmp = a[c] + i
                            if tmp > self.maxconstant+1:
                                 tmp = self.maxconstant+1
                            a[c] = tmp
                        #for r in resets: #This is wrong collected wrong data
                        #    a[r] = 0
                        exec update in self.constants, a 
                        #if normalize:
                        #    s1, w1, p1 = self.normalize(a,0,self.infconstant)
                        #else:
                        s1, w1, p1 = a,0,self.infconstant
                        #logger.debug("%s - %s - %s", s1, w1, p1)
                        yield s1, w1, p1 

#		#Check the guard using the values
#                if guard == None or self.satesfyguard(state, lowerVal, higherVal):
#                    #If synchronisation, find trans to sync with
#                    if sync != '':
#			print "not supported" #raise Exception("NotSupported")
#                    else:
#                        a = state.copy()
#                        a.locs[t_idx] = target_idx
#                        #Execute update
#                        logger.debug("Executing update: %s on %s: ", debug_info['update_code'], a)
#                        if update != None:
#                            try:
#                                exec update in self.constants, a
#                            except Exception, e:
#                                raise VirtualMachineException('Executing "' + debug_info['update_code'] + '": ' + str(e))
#                        #print "Result:", a.vars
#                        yield trans_info and (a, "") or a
    
    def delay_successors_1step(self, state):
        """Generate the 1 time unit delay successor"""
        a = state.copy()
        delay_has_effect = False
        for (c, cmax) in self.clocks:
            if a.clocks[c] <= cmax:
                a.clocks[c] = a.clocks[c]+1
                delay_has_effect = True
        if delay_has_effect:
            yield a

    def delay_successors_allsteps(self, state):
        """Generate all delay successors"""
        a = state.copy()
        delay_has_effect = True
        #Using invariants we might shortcut this loop when 
        # we see at state that violates the invariant as all guards are on the
        # form < and <= 
        while delay_has_effect:
            delay_has_effect = False
            for (c, cmax) in self.clocks:
                if a.clocks[c] <= cmax:
                    a.clocks[c] = a.clocks[c]+1
                    delay_has_effect = True
            if delay_has_effect:
                yield a
                a = a.copy()

    def delay_successors_interesting(self, state):
        """Generate the next "interesting" delay successor, that is the next
        state that has an enabled transition."""
        a = state.copy()
        delay_has_effect = True
        while delay_has_effect:
            delay_has_effect = False
            for (c, cmax) in self.clocks:
                if a.clocks[c] <= cmax:
                    a.clocks[c] = a.clocks[c]+1
                    delay_has_effect = True
            if delay_has_effect:
                #Check if delay is only option in new state
                for b in self.trans_successors(a):
                    yield a
                    #There was an option that was not delay,
                    #Stop generating delay successors
                    return

    def successors(self, state, w, p):
        for a in self.trans_successors(state, w, p):
            #if self.checkInvariant(a):
            yield a
        #for a in self.delay_successors_1step(state):
        #    if self.checkInvariant(a):
        #        yield a

    def successors_transinfo(self, state):
        for (a, ti) in self.trans_successors(state, trans_info=True):
            if self.checkInvariant(a):
                yield (a, ti)
        for a in self.delay_successors_1step(state):
            if self.checkInvariant(a):
                yield (a, "")

    def get_initialstate(self):
        #calculate initial state
        initstate = State(self)
        for (t, idx) in zip(self.model.templates, 
                            range(len(self.model.templates))):
            #Get initial location
            initstate.locs[idx] = t.locations.index(t.initlocation)
            #Set initial vars
            for (var, vartype, array_dimensions) in self.variables:
                if vartype in self.externs:
                    #assert len(array_dimensions) <= 1
                    if array_dimensions == []:
                        initval = self.externs[vartype]()
                    elif len(array_dimensions) == 1: #one dimension handled by lattice classes themselves
                        initval = self.externs[vartype](array_dimensions[0])
                    else:
                        #handle array dimensions, using LatticeVector class
                        #one dimension handled by lattice classes themselves
                        latticebase = self.externs[vartype](array_dimensions[0])
                        curval = latticebase
                        for dim in reversed(array_dimensions[1:]):
                            new_val = LatticeVector(curval, dim)
                            curval = new_val
                        initval = curval
                        
                    initstate.lattice_part[var] = initval
                    continue
                else:
                    try:
                        initval = {'bool': False, 'int': 0}[vartype]
                    except:
                        raise util.UnknownVarTypeException('Unknown Vartype: ' + vartype + ' ' + var)

                if array_dimensions == []:
                    initstate[var] = initval
                else:
                    curval = initval
                    for dim in reversed(array_dimensions):
                        new_val = []
                        for i in range(dim):
                            new_val.append(copy.deepcopy(curval))
                        curval = new_val
                    initstate[var] = curval

            #Set initial clocks
            for (c, cmax) in self.clocks:
                initstate.clocks[c] = 0

        logger.info("Initial state: locs: %s vars: %s", initstate.locs, initstate.vars)
        return initstate

# vim:ts=4:sw=4:expandtab
